use tokio::sync::{mpsc, oneshot};
use std::collections::HashMap;

/// The buffer size of the channel used to communicate with the dispatch actor.
const ACTOR_MESSAGE_BUFFER: usize = 1024;
/// The buffer size of the channels used to dispatch messages
const HANDLER_MESSAGE_BUFFER: usize = 1024;

enum DispatchMessage {
    /// Instruct dispatcher to send a message to a specific address
    SendMessage {
        to: String,
        from: String,
        message: Vec<u8>
    },
    /// Instruct dispatcher to provide a receiver through which messages to 'address' will be received.
    RegisterHandler {
        address: String,
        respond_to: oneshot::Sender<mpsc::Receiver<(String, Vec<u8>)>>,
    }
}

/// The handle to the dispatch actor, note the 'Clone' trait.
#[derive(Clone)]
pub struct DispatchHandle {
    sender: mpsc::Sender<DispatchMessage>
}

impl DispatchHandle {
    pub fn new() -> DispatchHandle {
        // Open a multiple sender, single consumer channel to communicate with the actor.
        let (sender, mut receiver) = mpsc::channel(ACTOR_MESSAGE_BUFFER);

        // Event loop for the dispatch actor, exits when all handles are dropped.
        tokio::spawn(async move {
            let mut actor = Actor::default();

            while let Some(message) = receiver.recv().await {
                actor.handle_message(message);
            }
        });

        // Return the handle
        DispatchHandle { sender }
    }

    /// Utility function to register a new address
    pub async fn register(&mut self, address: &str) -> mpsc::Receiver<(String, Vec<u8>)> {
        let (sender, receiver) = oneshot::channel();

        {
            let message = DispatchMessage::RegisterHandler { address: address.to_string(), respond_to: sender };

            self.sender.try_send(message).expect("Actor should be alive.");
        }

        receiver.await.expect("Channel should have been registered")
    }

    /// Utility function to send a message
    pub fn send_message(&mut self, to: &str, from: &str, message: Vec<u8>) {
        let message = DispatchMessage::SendMessage { to: to.to_string(), from: from.to_string(), message };

        self.sender.try_send(message).expect("Actor should be alive.");
    }
}

/// Represents a collection of channels that we can dispatch messages through
type DispatchChannels = Vec<mpsc::Sender<(String, Vec<u8>)>>;

/// The dispatch actor itself, that performs message routing.
#[derive(Default)]
struct Actor {
    senders: HashMap<String, DispatchChannels>
}

impl Actor {
    pub fn handle_message(&mut self, message: DispatchMessage) {
        match message {
            DispatchMessage::SendMessage { to, from, message} => {
                // Look for senders to dispatch message to
                if let Some(senders) = self.senders.get_mut(&to) {
                    senders
                        .iter_mut()
                        .try_for_each(|sender| sender.try_send((from.clone(), message.clone())))
                        .expect("Expected sender channel to be open and not full");
                } else {
                    eprintln!("No sender with address {} to dispatch - Message dropped.", to);
                }
            },
            DispatchMessage::RegisterHandler { address, respond_to } => {
                // Create a new channel
                let (sender, receiver) = mpsc::channel(HANDLER_MESSAGE_BUFFER);

                // Place sender in 'senders'
                self.senders.entry(address).or_default().push(sender);

                // Send back the new receiver
                respond_to.send(receiver).expect("Return channel has closed prematurely.");
            },
        }
    }
}